package gui;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.SwingConstants;
import javax.swing.WindowConstants;

import generic.Status;
import gui.panel.CommunicationPanel;
import gui.panel.ConnectionPanel;
import network.CommunicationManager;

public class GUI extends JFrame {
	private static final long serialVersionUID = 1L;
	private ConnectionPanel connectionPanel;
	private CommunicationPanel communicationPanel;
	private JPanel languagePanel;
	private JButton languageButton;
	private JTabbedPane tabbedPane;
	
	private CommunicationManager communicationManager;
	
	public GUI(CommunicationManager communicationManager) {
		super();
		this.communicationManager=communicationManager;
		connectionPanel=new ConnectionPanel(this);
		connectionPanel.setAddress(communicationManager.getAddress());
		connectionPanel.setPort(String.valueOf(communicationManager.getPort()));
		communicationPanel=new CommunicationPanel(this);
		initialize();
	}
	
	/**
	 * Initialize the default state of the different GUI elements.
	 */
	private void initialize() {
		this.setTitle("CommandPC - MARTIN Romain");
		this.setIconImage(getToolkit().getImage(getClass().getResource("/iconCommandPC.png")));
		this.setSize(600,300);
		this.setResizable(true);
		
		this.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
		this.addWindowListener(new java.awt.event.WindowAdapter() {
			@Override
			public void windowClosing(java.awt.event.WindowEvent e) {
				communicationManager.disconnect();
				dispose();
			};
		});
		
		GridBagLayout gbl_contentPane=new GridBagLayout();
		getContentPane().setLayout(gbl_contentPane);
		
		tabbedPane=new JTabbedPane(SwingConstants.TOP);
		tabbedPane.add("panelConnection",connectionPanel);
		tabbedPane.add("panelCommunication",communicationPanel);
		tabbedPane.setIconAt(0,new ImageIcon(getClass().getResource("/iconConnection.png")));
		tabbedPane.setIconAt(1,new ImageIcon(getClass().getResource("/iconCommunication.png")));
		tabbedPane.setEnabledAt(1,false);
		GridBagConstraints gbc_tabbedPane=new GridBagConstraints();
		gbc_tabbedPane.gridx=0;
		gbc_tabbedPane.gridy=0;
		gbc_tabbedPane.weightx=1.0d;
		gbc_tabbedPane.weighty=1.0d;
		gbc_tabbedPane.fill=GridBagConstraints.BOTH;
		getContentPane().add(tabbedPane,gbc_tabbedPane);
		
		GridBagLayout gbl_languagePanel=new GridBagLayout();
		languagePanel=new JPanel(gbl_languagePanel);
		GridBagConstraints gbc_languagePanel=new GridBagConstraints();
		gbc_languagePanel.gridx=0;
		gbc_languagePanel.gridy=1;
		gbc_languagePanel.fill=GridBagConstraints.HORIZONTAL;
		
		languageButton=new JButton("languageButton");
		languageButton.addActionListener(new java.awt.event.ActionListener() {
			@Override
			public void actionPerformed(java.awt.event.ActionEvent e) {
				if(Locale.getDefault().getLanguage().equals(Locale.FRENCH.toString())) {
					Locale.setDefault(Locale.ENGLISH);
				}else {
					Locale.setDefault(Locale.FRENCH);
				}
				//ResourceBundle.clearCache();
				setText();
			}
		});
		languagePanel.add(languageButton);
		
		getContentPane().add(languagePanel,gbc_languagePanel);
				
		setText();
	}
	
	/**
	 * Set the correct text for all graphical elements according to the default {@link Locale}.
	 * @see java.util.Locale#getDefault()
	 */
	private void setText(){
		ResourceBundle resourceBundle=ResourceBundle.getBundle("bundle/Bundle",Locale.getDefault());
		languageButton.setText(resourceBundle.getString("languageAlt"));
		tabbedPane.setTitleAt(0,resourceBundle.getString("connection"));
		tabbedPane.setTitleAt(1,resourceBundle.getString("command"));
		
		communicationPanel.setText();
		connectionPanel.setText();
	}

	/**
	 * Request connection status change.
	 * @param isConnection true if this is a connection request, false to disconnect.
	 * @param address address of the remote server.
	 * @param port port of the remote server.
	 * @return true if operation successful.
	 */
	public boolean changeConnectionStatus(boolean isConnection,String address,String port) {
		//Disconnection
		if(isConnection==false) {
			communicationManager.disconnect();
			tabbedPane.setEnabledAt(1,false);
			tabbedPane.setSelectedIndex(0);
			return true;
		}
		
		ResourceBundle resourceBundle=ResourceBundle.getBundle("bundle/Bundle",Locale.getDefault());

		if(address.isBlank() || port.isBlank()) {
			connectionPanel.addLog(resourceBundle.getString("errorSpecifyAddressAndPort"),true);
			return false;
		}
		
		Status<Object,Exception> result=communicationManager.connect(address,Integer.valueOf(port));
		if(result.success) {
			connectionPanel.addLog(resourceBundle.getString("connectedTo")+" "+result.message.toString(),false);
			tabbedPane.setEnabledAt(1,true);
			tabbedPane.setSelectedIndex(1);
		} else {
			connectionPanel.addLog(result.error.toString(),true);			
		}
		
		return result.success;
	}
	
	/**
	 * Send command to the server and receive his answer.
	 * @param text command to send.
	 */
	public void sendCommand(String text) {
		Status<Object,Exception> resultSend=communicationManager.encryptAndSendString(text);
		if(resultSend.success==false) {
			communicationPanel.addLog(resultSend.error.toString(), true);
			return;
		}
		Status<String, Exception> resultReceive=communicationManager.receiveAndDecryptString();
		if(resultReceive.success) {
			if(resultReceive.payload.trim().startsWith("java")) {
				communicationPanel.addLog(resultReceive.payload.trim(), true);
			} else {
				communicationPanel.addLog(resultReceive.payload.trim(), false);
			}
		} else {
			communicationPanel.addLog(resultReceive.error.toString(), true);
		}
	}
}
