package network;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.security.NoSuchAlgorithmException;
import java.util.Base64;

import javax.crypto.Cipher;
import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

import generic.Status;

public class CommunicationManager {
	private Client client;
	/** Key used to initialize the connection. */
	private SecretKey masterKey;
	/** Communication key between the client and the server. */
	private SecretKey communicationKey;
	/** If <code>true</code> then the message will be send using the communicationKey. Otherwise, the masterKey will be used.*/
	private boolean useCommunicationKey;
	
	public CommunicationManager(Client client) {
		this.client=client;
		this.masterKey=new SecretKeySpec(new byte[]{-70,-45,-79,-32,-36,108,-117,-7,-97,56,46,-86,-35,-11,-35,-81}, "AES");
		this.useCommunicationKey=false;		
	}
	
	public String getAddress() {
		if(client!=null) return client.getServerAddress();
		return "";
	}
	public int getPort() {
		if(client!=null) return client.getServerPort();
		return 0;
	}
	
	/**
	 * Establish connection with the server set in the {@link #client}.
	 * If no error, then swap the key from master to communication.
	 * @return response {@link Status} object.
	 */
	public Status<Object,Exception> connect() {
		Status<Object,Exception> status=new Status<Object,Exception>(true);

		//Stop immediately if already connected
		if(client!=null && client.getSocket()!=null && client.getSocket().isConnected()) {
			return status;
		}
		
		//Connect to server
		status=client.connect();
		if(status.success==false) {
			return status;
		}
		
		//Change from master key to a communication key
		Status<String,Exception> keyStatus=client.receiveString();
		if(keyStatus.success==false) {
			status.success=keyStatus.success;
			status.error=keyStatus.error;
			return status;
		}
		Status<String,Exception> decryptStatus=decryptString(keyStatus.payload, masterKey);
		if(decryptStatus.success==false) {
			status.success=decryptStatus.success;
			status.error=decryptStatus.error;
			return status;
		}
		String[] encodedKeyString=decryptStatus.payload.split(";");
		byte[] encodedKey=new byte[encodedKeyString.length];
		for(int i=0;i<encodedKeyString.length;i++) {
			encodedKey[i]=Byte.parseByte(encodedKeyString[i]);
		}
		communicationKey=new SecretKeySpec(encodedKey, "AES");
		useCommunicationKey=true;
		
		return status;		
	}
	
	/**
	 * Establish connection with a server.
	 * @param address server address
	 * @param port server remote port
	 * @return response {@link Status} object.
	 */
	public Status<Object,Exception> connect(String address,int port) {
		if(client.getSocket()!=null && client.getSocket().isConnected()) {
			disconnect();
		}
		this.client=new Client(port, address);
		return connect();
	}
	
	/**
	 * Encrypt and send data.
	 * @param data data to send.
	 * @return response {@link Status} object.
	 */
	public Status<Object,Exception> encryptAndSendString(String data) {
		String encryptedMessage="";
		if(useCommunicationKey) {
			encryptedMessage=encryptString(data, communicationKey).payload;
		}else {
			encryptedMessage=encryptString(data, masterKey).payload;
		}
		return client.sendString(encryptedMessage);
	}
	
	/**
	 * Receive and decrypt data.
	 * @return response {@link Status} object.
	 */
	public Status<String,Exception> receiveAndDecryptString() {
		Status<String,Exception> status=client.receiveString();
		if(status.success==false) {
			status.success=status.success;
			status.error=status.error;
			return status;
		}
		
		if(useCommunicationKey) {
			status=decryptString(status.payload, communicationKey);
		}else {
			status=decryptString(status.payload, masterKey);
		}
		
		if(status.success==false) {
			status.success=status.success;
			status.error=status.error;
		}
		return status;
	}
	
	/**
	 * Close the communication socket.
	 * Inform the remote client of the deconnection.
	 * @return response {@link Status} object.
	 */
	public Status<Object,IOException> disconnect() {
		encryptAndSendString("STOPstopSTOP");
		return client.close();
	}
	
	/**
	 * Generate and setup a communication key to replace the master key for future communication.
	 * Send the new key to the client.
	 */
	public void setupCommunicationKey() {
		communicationKey=generateSecretKey();
		String encodedKey="";
		for(int i=0;i<communicationKey.getEncoded().length;i++) {
			encodedKey+=communicationKey.getEncoded()[i]+";";
		}
		encryptAndSendString(encodedKey);
		useCommunicationKey=true;
 	}
	
	/**
	 * Generates a secret key.
	 * @return secret key.
	 */
	private SecretKey generateSecretKey() {
		KeyGenerator keyGenerator = null;
		try {
			keyGenerator = KeyGenerator.getInstance("AES");
		} catch (NoSuchAlgorithmException e) {
			//Never trigger as per the javax.crypto.KeyGenerator documentation:
			//Every implementation of the Java platform is required to support
			//the following standard KeyGenerator algorithms with the keysizes
			//in parentheses: AES (128)
			e.printStackTrace();
		}
		keyGenerator.init(128);
		return keyGenerator.generateKey();
	}
	
	/**
	 * Encrypt a message.
	 * @param data message to encrypt.
	 * @return response {@link Status} object.
	 */
	public Status<byte[],Exception> encrypt(byte[] data) {
		if(useCommunicationKey) {
			return encrypt(data,communicationKey);
		}else{
			return encrypt(data,masterKey);
		}
	}
	
	/**
	 * Encrypt a message with the specified key.
	 * @param data message to encrypt.
	 * @param secretKey secret key to use.
	 * @return response {@link Status} object.
	 */
	private Status<byte[],Exception> encrypt(byte[] data,SecretKey secretKey) {
		Status<byte[],Exception> result=new Status<byte[],Exception>(true);
		try {
			Cipher cipher = Cipher.getInstance("AES/ECB/PKCS5Padding");
			cipher.init(Cipher.ENCRYPT_MODE, secretKey);
			
			byte[] encrypt=cipher.doFinal(data);
			/*
			 * Base64.getEncoder().encode() is here to
			 * prevent javax.crypto.IllegalBlockSizeException:
			 * Input length must be multiple of 16 when decrypting with padded cipher
			 */
			result.payload=Base64.getEncoder().encode(encrypt);
		} catch (Exception e) {
			result.success=false;
			result.error=e;
		}
		return result;
	}
	
	/**
	 * Encrypt a message.
	 * @param data message to encrypt.
	 * @return response {@link Status} object.
	 */
	public Status<String,Exception> encryptString(String data) {
		if(useCommunicationKey) {
			return encryptString(data,communicationKey);
		}else{
			return encryptString(data,masterKey);
		}
	}
	
	/**
	 * Use {@link #encrypt(String, SecretKey)} to encrypt a message.
	 * @param data message to encrypt.
	 * @param secretKey secret key to use.
	 * @return response {@link Status} object.
	 */
	private Status<String,Exception> encryptString(String data,SecretKey secretKey) {
		Status<String,Exception> result=new Status<String,Exception>(true);
		Status<byte[],Exception> encryptResult=encrypt(data.getBytes(StandardCharsets.UTF_8),secretKey);
		
		result.message=encryptResult.message;
		result.success=encryptResult.success;
		result.error=encryptResult.error;
		if(encryptResult.success) {
			result.payload=new String(encryptResult.payload,StandardCharsets.UTF_8);
		}
		return result;
	}
	
	/*private String padStringForCrypto (String data) {
		//Prevent javax.crypto.IllegalBlockSizeException:
		//Input length must be multiple of 16 when decrypting with padded cipher
		if(data.length()%16!=0) {
			System.out.println("data length="+data.length());
			double dataLengthNeeded=Math.ceil((double)data.length()/16d)*16;
			data=String.format("%-"+(int)dataLengthNeeded+"s", data);
			System.out.println("data length="+data.length());
		}
		return data;
	}*/
		
	/**
	 * Decrypt a message.
	 * @param data encrypted message.
	 * @return response {@link Status} object.
	 */
	public Status<byte[],Exception> decrypt(byte[] data){
		if(useCommunicationKey) {
			return decrypt(data,communicationKey);
		}else{
			return decrypt(data,masterKey);
		}
	}
	
	/**
	 * Decrypt a message with the specified key.
	 * @param data encrypted message.
	 * @param secretKey secret key to use.
	 * @return response {@link Status} object.
	 */
	private Status<byte[],Exception> decrypt(byte[] data,SecretKey secretKey) {
		Status<byte[],Exception> result=new Status<byte[],Exception>(true);
		try {
			Cipher cipher = Cipher.getInstance("AES/ECB/PKCS5Padding");
			cipher.init(Cipher.DECRYPT_MODE, secretKey);
			/*
			 * Base64.getDecoder().decode() is here to
			 * prevent javax.crypto.IllegalBlockSizeException:
			 * Input length must be multiple of 16 when decrypting with padded cipher
			 */
			result.payload=cipher.doFinal(Base64.getDecoder().decode(data));
		} catch (Exception e) {
			result.success=false;
			result.error=e;
		}
		return result;
	}
	
	/**
	 * Decrypt a {@link String}.
	 * @param data encrypted message.
	 * @return response {@link Status} object.
	 */
	public Status<String,Exception> decryptString(String data) {
		if(useCommunicationKey) {
			return decryptString(data,communicationKey);
		}else{
			return decryptString(data,masterKey);
		}
	}
	
	/**
	 * Use {@link #encrypt(String, SecretKey)} to decrypt a message.
	 * @param data encrypted message.
	 * @param secretKey secret key to use.
	 * @return response {@link Status} object.
	 */
	private Status<String,Exception> decryptString(String data,SecretKey secretKey) {
		Status<String,Exception> result=new Status<String,Exception>(true);
		
		byte[] dataByte;
		try {
			dataByte=data.getBytes(StandardCharsets.UTF_8);
		} catch (NullPointerException e) {
			result.error=e;
			result.success=false;
			return result;
		}
		
		Status<byte[],Exception> decryptResult=decrypt(dataByte,secretKey);
		result.message=decryptResult.message;
		result.success=decryptResult.success;
		result.error=decryptResult.error;
		if(decryptResult.success) {
			result.payload=new String(decryptResult.payload,StandardCharsets.UTF_8);
		}
		
		return result;
	}
}
